@@ -19,6 +19,7 @@
 
 #include <gtk/gtk.h>
 #include <string.h>
+#include <stdio.h>
 
 #include "gpeiconlistview.h"
 #include "gpeiconlistitem.h"
@@ -27,7 +28,7 @@ static guint my_signals[2];
 
 struct _GPEIconListViewClass 
 {
-  GtkDrawingAreaClass parent_class;
+  GtkWidgetClass parent_class;
   void (* clicked) (GPEIconListView * self, gpointer udata);
   void (* show_popup) (GPEIconListView * self, gpointer udata);
 };
@@ -39,14 +40,13 @@ struct _Pack
   GPEIconListItem *icon;  //the icon if found, NULL if not found
 };
 
-static GtkDrawingAreaClass *parent_class;
+static GtkWidgetClass *parent_class;
 
 #define PARENT_HANDLER(___self,___allocation) \
 	{ if(GTK_WIDGET_CLASS(parent_class)->size_allocate) \
 		(* GTK_WIDGET_CLASS(parent_class)->size_allocate)(___self,___allocation); }
 
 
-static gint _gpe_icon_list_view_popup (gpointer data);
 static void _gpe_icon_list_view_cancel_popup (GPEIconListView *il);
 static GPEIconListItem *_gpe_icon_list_view_get_icon (GPEIconListView *il, int col, int row);
 
@@ -159,91 +159,79 @@ _gpe_icon_list_view_expose (GtkWidget *widget, GdkEventExpose *event)
 {
   GPEIconListView *il;
   GList *icons;
-  GdkPixbuf *dest;
   int row=0, col=0;
   PangoContext *pc;
   PangoLayout *pl;
   int label_height;
-  int i;
 	  
   il = GPE_ICON_LIST_VIEW (widget);
   
   /* Pango font rendering setup */
-  if ((pc = gtk_widget_get_pango_context (widget)) == NULL)
-    pc = gtk_widget_create_pango_context (widget);
-  
+  pc = gtk_widget_create_pango_context (widget);
+
   pl = pango_layout_new (pc);
   pango_layout_set_width (pl, il_col_width (il) * PANGO_SCALE);
   pango_layout_set_alignment (pl, PANGO_ALIGN_CENTER);
 	  
   label_height = il_label_height (il);
 
-  /* Make a new pixbuf for rendering icons to */
-  dest = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8,
-			 event->area.width, event->area.height);
-  
-  /* set the background color */
-  gdk_pixbuf_fill (dest, il->bgcolor);
-  
-  /* Paint the background image, scaled */
-  if (il->bgpixbuf)
-    {
-      /* ??? pb: what's this loop for?  why not just render the area of
-	 background image given by event->area?  */
-      for (i = 0; 
-	   i < il->rows * il_row_height (il); 
-	   i += gdk_pixbuf_get_height (il->bgpixbuf)) 
-	{
-	  GdkRectangle r1, dst;
-	  r1.x = 0; r1.y = i;
-	  r1.width = gdk_pixbuf_get_width (il->bgpixbuf);
-	  r1.height = gdk_pixbuf_get_height (il->bgpixbuf);
-	  
-	  if (gdk_rectangle_intersect (&r1, &(event->area), &dst)) {
-	    gdk_pixbuf_composite (il->bgpixbuf, dest,
-				  dst.x-event->area.x,
-				  dst.y-event->area.y,
-				  dst.width, dst.height,
-				  r1.x-event->area.x, r1.y-event->area.y,
-				  1,
-				  1,
-				  GDK_INTERP_BILINEAR, 255);
-	  }
-	}
-    }
-  
   for (icons = il->icons; icons != NULL; icons = icons->next) 
     {
       GdkRectangle r1, r2, dst;
       GPEIconListItem *icon;
-      GdkPixbuf *pixbuf=NULL;
-      
+      GdkPixbuf *pixbuf = NULL;
+      int cell_x, cell_y, cell_w, cell_h;
+ 
       icon = icons->data;
       
-      /* Compute & render the icon */
-      r1.x=col * il_col_width (il);
-      r1.y=row * il_row_height (il) + 5;
-      r1.width = il_col_width (il);
-      r1.height = il->icon_size;
-      
+      cell_x = col * il_col_width (il);
+      cell_y = row * il_row_height (il) + 5;
+      cell_w = il_col_width (il);
+      cell_h = il->icon_size + LABEL_YMARGIN + label_height;
+
       r2.x = event->area.x;
       r2.y = event->area.y;
       r2.width = event->area.width;
       r2.height = event->area.height;
-	      
-      if (gdk_rectangle_intersect (&r1, &r2, &dst)) 
+
+      r1.x = cell_x;
+      r1.y = cell_y;
+      r1.width = cell_w;
+      r1.height = cell_h;
+      
+      if (gdk_rectangle_intersect (&r1, &r2, &dst))
 	{
+	  gboolean selected = FALSE;
+	  GtkStateType state;
+
+	  if (il->mrow == row && il->mcol == col)
+	    selected = TRUE;
+
+	  state = (selected && !il->flag_embolden) ? GTK_STATE_SELECTED : GTK_WIDGET_STATE (widget);
+
+	  gtk_paint_flat_box (widget->style,
+			      widget->window,
+			      state,
+			      GTK_SHADOW_NONE,
+			      &dst,
+			      widget,
+			      "",
+			      cell_x,
+			      cell_y,
+			      cell_w,
+			      cell_h);
+	  
 	  /* Get the icon from the cache if its there, if not put it there :) */
 	  if (icon->pb_scaled)
 	    pixbuf = icon->pb_scaled;
 	  else 
 	    {
-	      if (icon->pb == NULL)
+	      if (icon->pb == NULL && icon->icon != NULL)
 		icon->pb = gdk_pixbuf_new_from_file (icon->icon, NULL);
 	      
 	      _gpe_icon_list_view_check_icon_size (il, G_OBJECT (icon));
 	      
-	      if (icon->icon) 
+	      if (icon->icon && icon->pb)
 		{
 		  gdk_pixbuf_unref (icon->pb);
 		  icon->pb = NULL;
@@ -252,126 +240,100 @@ _gpe_icon_list_view_expose (GtkWidget *widget, GdkEventExpose *event)
 	      pixbuf = icon->pb_scaled;
 	    }
 	  
-	  if (pixbuf) 
+	  if (pixbuf)
 	    {
+	      int pixbuf_w, pixbuf_h, x_offset;
+	      
+	      pixbuf_w = gdk_pixbuf_get_width (pixbuf);
+	      pixbuf_h = gdk_pixbuf_get_height (pixbuf);
+
 	      // adjust X position for actual size
-	      r1.width = gdk_pixbuf_get_width (pixbuf);
-	      r1.x=col * il_col_width (il) + (il_col_width (il) - r1.width) / 2;
+	      x_offset = (il_col_width (il) - pixbuf_w) / 2;
+
+	      r1.x = cell_x + x_offset;
+	      r1.y = cell_y;
+	      r1.width = pixbuf_w;
+	      r1.height = pixbuf_h;
+	      
+	      if (gdk_rectangle_intersect (&r1, &r2, &dst))
+		gdk_draw_pixbuf (widget->window, 
+				 widget->style->fg_gc[widget->state],
+				 pixbuf, 
+				 dst.x - r1.x, dst.y - r1.y, 
+				 dst.x, dst.y, 
+				 dst.width, dst.height,
+				 GDK_RGB_DITHER_NORMAL, 0, 0);
+	    }
+	  
+	  if (il->flag_show_title && icon->title)
+	    {
+	      /* if show_title mode is set and current title non NULL, then */
+	      /* Compute & render the title */
+	      r1.x = cell_x;
+	      r1.y = row * il_row_height (il) + il->icon_size + LABEL_YMARGIN;
+	      r1.width = cell_w;
+	      r1.height = label_height;
 	      
-	      // recompute intersection: maybe nothing to do any more
 	      if (gdk_rectangle_intersect (&r1, &r2, &dst)) 
 		{
-		  gdk_pixbuf_composite (pixbuf, dest, // from, to
-					dst.x - event->area.x, //dest_x
-					dst.y - event->area.y,// + va->value, // dest_y
-					dst.width, dst.height, // dest_width, dest_height
-					r1.x - event->area.x, // offset_x
-					r1.y - event->area.y, // + va->value, // offset_y
-					1.0, 1.0,
-					GDK_INTERP_BILINEAR, // filtering
-					col == il->mcol && row == il->mrow ? 128 : 255);
+		  char *stxt; int slen;
+		  PangoRectangle pr;
+		  
+		  stxt=icon->title;
+		  slen = strlen (stxt);
+		  
+		  if (selected && il->flag_embolden) 
+		    {
+		      char *newtxt;
+		      newtxt = g_strdup_printf ("<b>%s</b>", stxt);
+		      pango_layout_set_markup (pl, newtxt, -1);
+		      g_free (newtxt);
+		    } 
+		  else
+		    pango_layout_set_text (pl, stxt, -1);
+		  
+		  pango_layout_get_pixel_extents  (pl, NULL, &pr);
+	      
+		  while ((pango_layout_get_line_count (pl) > 2 ||
+			  pr.width > il_col_width (il)) && slen>0) {
+		    char *newtxt;
+		    char *fmt;
+		    if (stxt[slen-2] == ' ')
+		      slen--;
+		    
+		    if (selected && il->flag_embolden)
+		      fmt = g_strdup_printf ("<b>%%.%ds...</b>", --slen);
+		    else
+		      fmt = g_strdup_printf ("%%.%ds...", --slen);
+		    
+		    newtxt = g_strdup_printf (fmt, stxt);
+		    g_free (fmt);
+		    
+		    if (selected)
+		      pango_layout_set_markup (pl, newtxt, -1);
+		    else
+		      pango_layout_set_text (pl, newtxt, -1);
+		    g_free (newtxt);
+		    
+		    pango_layout_get_pixel_extents  (pl, NULL, &pr);
+		  }
+		  
+		  gtk_paint_layout (widget->style, widget->window, state,
+				    FALSE, &dst, widget, "", r1.x, r1.y, pl);
 		}
-	    }
+	    } //(if show title, compute and render)
 	}
-      
+
       if (++col == il->cols) 
 	{
 	  col = 0; 
 	  row++;
 	}
     }
-  
-  /* Dump to drawingarea */
-  gdk_pixbuf_render_to_drawable (dest, widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
-				 0, 0, event->area.x, event->area.y,
-				 event->area.width, event->area.height,
-				 GDK_RGB_DITHER_NORMAL, event->area.x, event->area.y);
-  
-  gdk_pixbuf_unref (dest);
-  
-  col = row = 0;
-  
-  for (icons = il->icons; icons != NULL; icons = icons->next) 
-    {
-      GdkRectangle r1, r2, dst;
-      GPEIconListItem *icon;
-      icon = icons->data;
-      
-      if (il->flag_show_title && icon->title)
-	{
-	  /* if show_title mode is set and current title non NULL, then */
-	  /* Compute & render the title */
-	  r1.x=col * il_col_width (il);
-	  r1.y=row * il_row_height (il) + il->icon_size + LABEL_YMARGIN;
-	  r1.width = il_col_width (il);
-	  r1.height = label_height;
-	  
-	  r2.x = event->area.x;
-	  r2.y = event->area.y;
-	  r2.width = event->area.width;
-	  r2.height = event->area.height;
-	  
-	  if (gdk_rectangle_intersect (&r1, &r2, &dst)) 
-	    {
-	      char *stxt; int slen;
-	      int selected=0;
-	      PangoRectangle pr;
-	      
-	      stxt=icon->title;
-	      slen = strlen (stxt);
-	      
-	      selected = (il->mrow == row && il->mcol == col);
-	      if (selected && il->flag_embolden) 
-		{
-		  char *newtxt;
-		  newtxt = g_strdup_printf ("<b>%s</b>", stxt);
-		  pango_layout_set_markup (pl, newtxt, -1);
-		  g_free (newtxt);
-		} 
-	      else
-		pango_layout_set_text (pl, stxt, -1);
-	      
-	      pango_layout_get_pixel_extents  (pl, NULL, &pr);
-	      
-	      while ((pango_layout_get_line_count (pl) > 2 ||
-		      pr.width > il_col_width (il)) && slen>0) {
-		char *newtxt;
-		char *fmt;
-		if (stxt[slen-2] == ' ')
-		  slen--;
-		
-		if (selected && il->flag_embolden)
-		  fmt = g_strdup_printf ("<b>%%.%ds...</b>", --slen);
-		else
-		  fmt = g_strdup_printf ("%%.%ds...", --slen);
-		
-		newtxt = g_strdup_printf (fmt, stxt);
-		g_free (fmt);
-		
-		if (selected)
-		  pango_layout_set_markup (pl, newtxt, -1);
-		else
-		  pango_layout_set_text (pl, newtxt, -1);
-		g_free (newtxt);
-		
-		pango_layout_get_pixel_extents  (pl, NULL, &pr);
-	      }
-	      
-	      gtk_paint_layout (widget->style, widget->window, GTK_WIDGET_STATE (widget),
-				FALSE, &dst,
-				widget, "detail?wtf?", r1.x, r1.y, pl);
-	    }
-	} //(if show title, compute and render)
-      
-      if (++col == il->cols) 
-	{
-	  col = 0; 
-	  row++;
-	}
-    }
-  
+
   g_object_unref (pl);
-  
+  g_object_unref (pc);
+
   return TRUE;
 }
 
@@ -382,17 +344,22 @@ _gpe_icon_list_view_button_press (GtkWidget *widget, GdkEventButton *event)
   GPEIconListItem *data;
   
   il = GPE_ICON_LIST_VIEW (widget);
-  
+
   _gpe_icon_list_view_get_rowcol (il, event->x, event->y, &(il->mcol), &(il->mrow));
   _gpe_icon_list_view_refresh_containing (il, il->mcol, il->mrow);
   
-  data = _gpe_icon_list_view_get_icon(il, il->mcol, il->mrow);
-  if (data) 
+  data = _gpe_icon_list_view_get_icon (il, il->mcol, il->mrow);
+
+  switch (event->button)
     {
-      gpe_icon_list_item_button_press (data, event);
-      
-      /* Register a popup if there is an icon under the cursor */
-      il->popup_timeout = gtk_timeout_add (500, _gpe_icon_list_view_popup, (gpointer)il);
+    case 1:
+      if (data)
+	gpe_icon_list_item_button_press (data, event);
+      break;
+
+    case 3:
+      g_signal_emit (G_OBJECT (il), my_signals[1], 0, data ? data->udata : NULL);
+      break;
     }
   
   return TRUE;
@@ -442,7 +409,6 @@ _gpe_icon_list_view_recalc_size (GPEIconListView *self, GtkAllocation *allocatio
   if (self->rows_set && self->rows > self->rows_set)
     self->rows = self->rows_set;
  
-  // update drawing area
   da_new_height = self->rows * il_row_height (self) + TOP_MARGIN;
   gtk_widget_set_usize (GTK_WIDGET (self), -1, da_new_height);
 }
@@ -568,20 +534,56 @@ gpe_icon_list_view_set_icon_size (GPEIconListView *self, guint size)
 }
 
 static void
-_gpe_icon_list_view_size_allocate (GtkWidget *self, GtkAllocation *allocation)
+_gpe_icon_list_view_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 {
-  GPEIconListView *il;
-  il = GPE_ICON_LIST_VIEW (self);
+  GPEIconListView *view;
+  gboolean size_change = TRUE;
+
+  view = GPE_ICON_LIST_VIEW (widget);
+
+  if (allocation->width == widget->allocation.width
+      && allocation->height == widget->allocation.height)
+    size_change = FALSE;
   
-  if ((int)g_object_get_data (G_OBJECT(il), "size_x") == allocation->width &&
-      (int)g_object_get_data (G_OBJECT(il), "size_y") == allocation->height)
-    return;
+  widget->allocation = *allocation;
+
+  if (GTK_WIDGET_REALIZED (widget))
+    gdk_window_move_resize (widget->window,
+			    allocation->x, allocation->y,
+			    allocation->width, allocation->height);
   
-  g_object_set_data (G_OBJECT(il), "size_x", (gpointer)allocation->width);
-  g_object_set_data (G_OBJECT(il), "size_y", (gpointer)allocation->height);
-  
-  PARENT_HANDLER (self, allocation);
-  _gpe_icon_list_view_recalc_size (il, &self->allocation);	//luc: recalc *after* the size is allocated :)
+  if (size_change)
+    _gpe_icon_list_view_recalc_size (view, &widget->allocation);  //luc: recalc *after* the size is allocated :)
+}
+
+static void
+_gpe_icon_list_view_realize (GtkWidget *widget)
+{
+  GdkWindowAttr attributes;
+  gint attributes_mask;
+
+  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
+    
+  attributes.window_type = GDK_WINDOW_CHILD;
+  attributes.x = widget->allocation.x;
+  attributes.y = widget->allocation.y;
+  attributes.width = widget->allocation.width;
+  attributes.height = widget->allocation.height;
+  attributes.wclass = GDK_INPUT_OUTPUT;
+  attributes.visual = gtk_widget_get_visual (widget);
+  attributes.colormap = gtk_widget_get_colormap (widget);
+  attributes.event_mask = gtk_widget_get_events (widget);
+  attributes.event_mask |= (GDK_EXPOSURE_MASK |
+			    GDK_BUTTON_PRESS_MASK |
+			    GDK_BUTTON_RELEASE_MASK);
+
+  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
+
+  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
+  gdk_window_set_user_data (widget->window, widget);
+
+  widget->style = gtk_style_attach (widget->style, widget->window);
+  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);
 }
 
 static GPEIconListItem *
@@ -595,7 +597,6 @@ _gpe_icon_list_view_get_icon (GPEIconListView *il, int col, int row)
     icons = icons->next;
   
   return icons ? icons->data : NULL;
-  
 }
 
 static void 
@@ -607,20 +608,6 @@ _gpe_icon_list_view_cancel_popup (GPEIconListView *il)
   }
 }
 
-static gint 
-_gpe_icon_list_view_popup (gpointer data) 
-{
-  GPEIconListView *il = data;
-  int row, col;
-  
-  row = il->mrow;
-  col = il->mcol;
-  
-  g_signal_emit (G_OBJECT (il), my_signals[1], 0, _gpe_icon_list_view_get_icon(il,col, row)->udata);
-  
-  return FALSE;
-}
-
 void 
 gpe_icon_list_view_popup_removed (GPEIconListView *self) 
 {
@@ -648,12 +635,10 @@ gpe_icon_list_view_init (GPEIconListView *self)
   self->icon_size = 48;
   self->icon_xmargin = 12;
   self->bgcolor = 0xffffffff;
-  self->flag_embolden = TRUE;
+  self->flag_embolden = FALSE;
   self->flag_show_title = TRUE;
   self->rows_set = 0;
 
-  gtk_widget_add_events (GTK_WIDGET (self), GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
-
   self->label_height = _gpe_icon_list_view_title_height (self);
 }
 
@@ -661,7 +646,7 @@ static void
 gpe_icon_list_view_class_init (GPEIconListViewClass * klass)
 {
   GtkWidgetClass *widget_class;
-  parent_class = g_type_class_ref (GTK_TYPE_DRAWING_AREA);
+  parent_class = g_type_class_ref (GTK_TYPE_WIDGET);
 
   my_signals[0] = g_signal_new ("clicked",
 				G_TYPE_FROM_CLASS (klass),
@@ -686,6 +671,7 @@ gpe_icon_list_view_class_init (GPEIconListViewClass * klass)
   widget_class->button_press_event = _gpe_icon_list_view_button_press;
   widget_class->button_release_event = _gpe_icon_list_view_button_release;
   widget_class->size_allocate = _gpe_icon_list_view_size_allocate;
+  widget_class->realize = _gpe_icon_list_view_realize;
 }
 
 static void
@@ -713,7 +699,7 @@ gpe_icon_list_view_get_type (void)
 	(GInstanceInitFunc) gpe_icon_list_view_init,
       };
 
-      type = g_type_register_static (GTK_TYPE_DRAWING_AREA, "GPEIconListView", &info, (GTypeFlags)0);
+      type = g_type_register_static (GTK_TYPE_WIDGET, "GPEIconListView", &info, (GTypeFlags)0);
     }
 
   return type;
